package org.bench4q.agent.scenario;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.bench4q.agent.datacollector.DataCollector;
import org.bench4q.agent.helper.ApplicationContextHelper;
import org.bench4q.agent.parameterization.SessionObject;
import org.bench4q.agent.plugin.Plugin;
import org.bench4q.agent.plugin.PluginManager;
import org.bench4q.agent.plugin.result.HttpReturn;
import org.bench4q.agent.plugin.result.PluginReturn;
import org.bench4q.agent.scenario.behavior.Behavior;

public class Worker implements Runnable {
	private ScenarioContext scenarioContext;
	private SessionObject sessionObject;

	private PluginManager pluginManager;

	private ScenarioContext getScenarioContext() {
		return scenarioContext;
	}

	private void setScenarioContext(ScenarioContext scenarioContext) {
		this.scenarioContext = scenarioContext;
	}

	private SessionObject getSessionObject() {
		return sessionObject;
	}

	private void setSessionObject(SessionObject sessionObject) {
		this.sessionObject = sessionObject;
	}

	private PluginManager getPluginManager() {
		return pluginManager;
	}

	private void setPluginManager(PluginManager pluginManager) {
		this.pluginManager = pluginManager;
	}

	public Worker(ScenarioContext scenarioContext, SessionObject sessionObject) {
		this.setScenarioContext(scenarioContext);
		this.setPluginManager(ApplicationContextHelper.getContext().getBean(
				PluginManager.class));
		this.setSessionObject(sessionObject);
	}

	public void run() {
		doRunScenario(getScenarioContext());
		doCleanUp();
	}

	private void doCleanUp() {
		this.sessionObject.doCleanUp();
	}

	public void doRunScenario(ScenarioContext context) {
		Map<String, Object> plugins = new HashMap<String, Object>();
		preparePlugins(context.getScenario(), plugins);
		for (int i = 0; i < context.getScenario().getPages().length; i++) {
			Page page = context.getScenario().getPages()[i];
			context.getDataStatistics().add(
					PageResult.buildPageResult(
							i,
							doRunBatchesInPage(plugins, page,
									context.getDataStatistics())));
		}

	}

	private List<BehaviorResult> doRunBatchesInPage(
			Map<String, Object> plugins, Page page, DataCollector dataCollector) {
		List<BehaviorResult> results = new ArrayList<BehaviorResult>();
		for (Batch batch : page.getBatches()) {
			results.addAll(doRunBatch(plugins, batch, dataCollector));
		}
		return results;
	}

	private List<BehaviorResult> doRunBatch(Map<String, Object> plugins,
			Batch batch, DataCollector dataCollector) {
		List<BehaviorResult> results = new ArrayList<BehaviorResult>();
		for (Behavior behavior : batch.getBehaviors()) {
			// each execution should call this
			behavior.distillParams(this.getSessionObject());
			Object plugin = plugins.get(behavior.getUse());
			Map<String, String> behaviorParameters = prepareBehaviorParameters(behavior);
			Date startDate = new Date(System.currentTimeMillis());
			PluginReturn pluginReturn = (PluginReturn) this.getPluginManager()
					.doBehavior(plugin, behavior.getName(), behaviorParameters);
			extractRunTimeParams(pluginReturn);
			Date endDate = new Date(System.currentTimeMillis());
			if (!behavior.shouldBeCountResponseTime()) {
				continue;
			}
			BehaviorResult behaviorResult = buildBehaviorResult(behavior,
					plugin, startDate, pluginReturn, endDate);
			pluginReturn = null;
			dataCollector.add(behaviorResult);
			results.add(behaviorResult);
		}
		return results;
	}

	private void extractRunTimeParams(PluginReturn pluginReturn) {
		this.getSessionObject().saveRuntimeParams(
				pluginReturn.getRunTimeParams());
	}

	private BehaviorResult buildBehaviorResult(Behavior behavior,
			Object plugin, Date startDate, PluginReturn pluginReturn,
			Date endDate) {
		BehaviorResult result = new BehaviorResult();
		result.setStartDate(startDate);
		result.setEndDate(endDate);
		result.setSuccess(pluginReturn.isSuccess());
		result.setResponseTime(endDate.getTime() - startDate.getTime());
		result.setBehaviorName(behavior.getName());
		result.setPluginId(behavior.getUse());
		result.setPluginName(plugin.getClass().getAnnotation(Plugin.class)
				.value());
		result.setShouldBeCountResponseTime(behavior
				.shouldBeCountResponseTime());
		if (pluginReturn instanceof HttpReturn) {
			HttpReturn httpReturn = (HttpReturn) pluginReturn;
			result.setBehaviorId(behavior.getId());
			for (Parameter parameter : behavior.getParameters()) {
				if (parameter.getKey().equals("url")) {
					result.setBehaviorUrl(parameter.getValue());
				}
			}
			result.setContentLength(httpReturn.getContentLength());
			result.setContentType(httpReturn.getContentType());
			result.setStatusCode(httpReturn.getStatusCode());
		}
		return result;
	}

	private Map<String, String> prepareBehaviorParameters(Behavior behavior) {
		Map<String, String> behaviorParameters = new HashMap<String, String>();
		for (Parameter parameter : behavior.getParameters()) {
			behaviorParameters.put(parameter.getKey(), parameter.getValue());
		}
		return behaviorParameters;
	}

	private void preparePlugins(Scenario scenario, Map<String, Object> plugins) {
		for (UsePlugin usePlugin : scenario.getUsePlugins()) {
			String pluginId = usePlugin.getId();
			Class<?> pluginClass = this.getPluginManager().getPlugins()
					.get(usePlugin.getName());
			Map<String, String> initParameters = new HashMap<String, String>();
			for (Parameter parameter : usePlugin.getParameters()) {
				initParameters.put(parameter.getKey(), parameter.getValue());
			}
			Object plugin = this.getPluginManager().initializePlugin(
					pluginClass, initParameters);
			plugins.put(pluginId, plugin);
		}
	}
}
