package org.bench4q.master.domain.service;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.apache.log4j.Logger;
import org.bench4q.master.Main;
import org.bench4q.master.domain.entity.TestPlan;
import org.bench4q.master.domain.entity.User;
import org.bench4q.master.domain.repository.TestPlanRepository;
import org.bench4q.master.testplan.highavailable.CurrentLoadObserver;
import org.bench4q.master.testplan.highavailable.HighAvailablePool;
import org.bench4q.master.testplan.schedulscript.TaskCompleteCallback;
import org.bench4q.share.enums.master.TestPlanStatus;
import org.bench4q.share.models.master.TestPlanModel;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.Restrictions;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

@Component
public class TestPlanEngine implements TaskCompleteCallback,
		CurrentLoadObserver {
	private TestPlanService testPlanService;
	private HighAvailablePool haPool;
	private TestPlanRepository testPlanRepository;
	private ScheduledExecutorService scheduleExecutor;

	public static Logger logger = Logger.getLogger(TestPlanEngine.class);

	private TestPlanRepository getTestPlanRepository() {
		return testPlanRepository;
	}

	@Autowired
	private void setTestPlanRepository(TestPlanRepository testPlanRepository) {
		this.testPlanRepository = testPlanRepository;
	}

	private TestPlanService getTestPlanService() {
		return testPlanService;
	}

	@Autowired
	private void setTestPlanService(TestPlanService testPlanService) {
		this.testPlanService = testPlanService;
	}

	public HighAvailablePool getHaPool() {
		return haPool;
	}

	@Autowired
	private void setHaPool(HighAvailablePool haPool) {
		this.haPool = haPool;
		this.getHaPool().setObserver(this);
	}

	private ScheduledExecutorService getScheduleExecutor() {
		return scheduleExecutor;
	}

	private void setScheduleExecutor(ScheduledExecutorService scheduleExecutor) {
		this.scheduleExecutor = scheduleExecutor;
	}

	public TestPlanEngine() {
		init();
	}

	private void init() {
		this.setScheduleExecutor(new ScheduledThreadPoolExecutor(1));
		this.getScheduleExecutor().scheduleAtFixedRate(new Runnable() {
			public void run() {
				TestPlan testPlan = pickATestPlan();

				if (testPlan == null) {
					logger.info("testPlan by pick is null");
					return;
				}
				logger.info(testPlan.getTestPlanRunId() + " start to run");
				doRunTestPlan(UUID.fromString(testPlan.getTestPlanRunId()));
			}
		}, 0, Main.PICK_CYCLE_IN_SECONDS, TimeUnit.SECONDS);
	}

	public UUID runWith(final TestPlanModel testPlanModel, User user) {
		ExecutorService executorService = Executors.newCachedThreadPool();
		final UUID testPlanId = UUID.randomUUID();

		if (!this.getTestPlanService().submitTestPlan(testPlanModel, user,
				testPlanId)) {
			return null;
		}
		Runnable runnable = new Runnable() {
			public void run() {
				doRunTestPlan(testPlanId);
			}
		};
		executorService.execute(runnable);
		executorService.shutdown();
		return testPlanId;
	}

	public void doRunTestPlan(final UUID testPlanId) {
		TestPlan testPlan = getTestPlanRepository().getTestPlanBy(testPlanId);
		TestPlanStatus returnStatus = testPlan.run();
		commitTestPlanStaus(returnStatus, testPlan);
	}

	private void commitTestPlanStaus(TestPlanStatus returnStatus,
			TestPlan testPlan) {
		if (returnStatus == TestPlanStatus.InRunning) {
			commitInRunning(testPlan);
		} else {
			commitError(testPlan, returnStatus);
		}
	}

	private void commitError(TestPlan testPlan, TestPlanStatus errorStatus) {
		testPlan.setCurrentStatus(errorStatus.name());
		getTestPlanRepository().updateEntity(testPlan);
	}

	private void commitInRunning(TestPlan testPlan) {
		testPlan.setCurrentStatus(TestPlanStatus.InRunning.name());
		testPlan.setLastRunningTime(new Date());
		getTestPlanRepository().attachRunningTestPlan(testPlan);
		getTestPlanRepository().updateEntity(testPlan);
	}

	public void doTaskComplete(UUID testPlanID) {
		TestPlan testPlan = this.getTestPlanRepository().getRunningTestPlanBy(
				testPlanID);
		if (testPlan == null || !testPlan.isFinish()) {
			return;
		}
		this.getHaPool().doForTestPlanComplete(testPlan.getTestPlanScripts());
		commitComplete(testPlan);
		this.getTestPlanRepository().detachRunningTestPlan(testPlanID);
		logger.info("Test plan with id " + testPlanID.toString() + " finishes!");
	}

	private void commitComplete(TestPlan testPlan) {
		testPlan.setCurrentStatus(TestPlanStatus.Complete.name());
		this.getTestPlanRepository().updateEntity(testPlan);
	}

	public void executePendingTestPlan() {

	}

	public TestPlan pickATestPlan() {
		logger.info("enter pick");
		List<Criterion> criterions = new ArrayList<Criterion>();
		Criterion notComplete = Restrictions.not(Restrictions.eq(
				"currentStatus", TestPlanStatus.Complete.name()));
		criterions.add(notComplete);
		Criterion notInRunning = Restrictions.not(Restrictions.eq(
				"currentStatus", TestPlanStatus.InRunning.name()));
		criterions.add(notInRunning);
		Criterion lessThanPoolAvailbleLoad = Restrictions.lt("requiredLoad",
				this.getHaPool().getCurrentAvailableLoad());
		criterions.add(lessThanPoolAvailbleLoad);
		Criterion lessThanMaxFailTimes = Restrictions.lt("failTimes",
				Main.MAX_FAIL_TIMES);
		criterions.add(lessThanMaxFailTimes);
		Criterion intervalGTMin_Interval = Restrictions.lt("lastRunningTime",
				new Date(System.currentTimeMillis()
						- Main.MIN_EXECUTE_INTERVAL_IN_SECONDS * 1000));
		criterions.add(intervalGTMin_Interval);
		List<TestPlan> testPlans = this.getTestPlanRepository()
				.loadTestPlansBy(criterions, Order.desc("createDateTime"));
		return testPlans.size() > 0 ? testPlans.get(0) : null;
	}
}
