package org.bench4q.master.infrastructure.communication;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.UUID;

import javax.xml.bind.JAXBException;

import org.apache.log4j.Logger;
import org.bench4q.master.domain.entity.Agent;
import org.bench4q.master.exception.ExceptionLog;
import org.bench4q.master.testplan.highavailable.faultolerence.FaultTolerenceFactory;
import org.bench4q.share.communication.HttpRequester;
import org.bench4q.share.communication.HttpRequester.HttpResponse;
import org.bench4q.share.helper.MarshalHelper;
import org.bench4q.share.models.agent.RunScenarioModel;
import org.bench4q.share.models.agent.RunScenarioResultModel;
import org.bench4q.share.models.agent.ServerStatusModel;
import org.bench4q.share.models.agent.StopTestModel;
import org.bench4q.share.models.agent.statistics.AgentBriefStatusModel;
import org.bench4q.share.models.agent.statistics.AgentBehaviorsBriefModel;
import org.bench4q.share.models.agent.statistics.AgentPageBriefModel;
import org.bench4q.share.models.agent.statistics.AgentPagesBriefModel;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

@Component
public class AgentMessenger {
	private HttpRequester httpRequester;
	private Logger logger = Logger.getLogger(AgentMessenger.class);

	private HttpRequester getHttpRequester() {
		return httpRequester;
	}

	@Autowired
	private void setHttpRequester(HttpRequester httpRequester) {
		this.httpRequester = httpRequester;
	}

	public RunScenarioResultModel runWithoutParams(Agent agent,
			RunScenarioModel runScenarioModel) throws IOException {
		HttpResponse httpResponse = null;
		try {
			httpResponse = this.httpRequester.sendPostXml(buildBaseUrl(agent)
					+ "/test/runWithoutParams", MarshalHelper.marshal(
					RunScenarioModel.class, runScenarioModel), null);
			if (HttpRequester.isInvalidResponse(httpResponse)) {
				return null;
			}
			return (RunScenarioResultModel) MarshalHelper.unmarshal(
					RunScenarioResultModel.class, httpResponse.getContent());
		} catch (JAXBException e) {
			logIt(httpResponse, e);
			FaultTolerenceFactory.getRunAgentFault(agent, runScenarioModel)
					.doTolerance();
			return null;
		}

	}

	private void logIt(HttpResponse httpResponse, Throwable e) {
		logger.error(HttpRequester.isInvalidResponse(httpResponse) ? ""
				: httpResponse.getContent(), e);
	}

	public RunScenarioResultModel submitScenrioWithParams(Agent agent,
			List<File> paramFiles, final RunScenarioModel runScenarioModel) {
		HttpResponse httpResponse = null;
		try {
			final String modelCOntent = MarshalHelper
					.tryMarshal(runScenarioModel);
			httpResponse = this.httpRequester.postFiles(buildBaseUrl(agent)
					+ "/test/submitScenarioWithParams", "files[]", paramFiles,
					"scenarioModel", new LinkedList<String>() {
						private static final long serialVersionUID = 1L;
						{
							add(modelCOntent);
						}
					});
			if (HttpRequester.isInvalidResponse(httpResponse)) {
				return null;
			}
			return (RunScenarioResultModel) MarshalHelper.unmarshal(
					RunScenarioResultModel.class, httpResponse.getContent());
		} catch (Exception e) {
			logIt(httpResponse, e);
			return null;
		}
	}

	public RunScenarioResultModel runWithParams(Agent agent, UUID agentRunId) {
		HttpResponse httpResponse = null;
		try {
			if (agent == null || agentRunId == null) {
				return null;
			}
			httpResponse = this.httpRequester.sendPost(buildBaseUrl(agent)
					+ "/test/runWithParams/" + agentRunId.toString(), null,
					null);
			if (HttpRequester.isInvalidResponse(httpResponse)) {
				logger.error("the response is not valid");
				return null;
			}
			return (RunScenarioResultModel) MarshalHelper.unmarshal(
					RunScenarioResultModel.class, httpResponse.getContent());
		} catch (Exception e) {
			logIt(httpResponse, e);
			return null;
		}
	}

	public AgentBriefStatusModel brief(Agent agent, UUID agentRunId) {
		HttpResponse httpResponse = null;
		try {
			if (agent == null || agentRunId == null) {
				return null;
			}
			httpResponse = this.httpRequester.sendGet(buildBaseUrl(agent)
					+ "/test/brief/" + agentRunId, null, null);
			if (HttpRequester.isInvalidResponse(httpResponse)) {
				return null;
			}
			logger.info(httpResponse.getContent());
			return (AgentBriefStatusModel) MarshalHelper.unmarshal(
					AgentBriefStatusModel.class, httpResponse.getContent());
		} catch (Exception e) {
			logIt(httpResponse, e);
			logger.error(e.toString() + " When brief the agent with hostName "
					+ agent.getHostName());
			FaultTolerenceFactory.getBriefFaultTolerance(agent).doTolerance();
			return null;
		}
	}

	public StopTestModel stop(Agent agent, UUID runId) {
		HttpResponse httpResponse;
		try {
			if (agent == null || runId == null) {
				return null;
			}
			httpResponse = this.httpRequester.sendGet(buildBaseUrl(agent)
					+ "/test/stop/" + runId.toString(), null, null);
			if (HttpRequester.isInvalidResponse(httpResponse)) {
				return null;
			}
			return (StopTestModel) MarshalHelper.unmarshal(StopTestModel.class,
					httpResponse.getContent());
		} catch (Exception e) {
			logger.error(e.toString() + " when stop the agent with hostName "
					+ agent.getHostName());
			FaultTolerenceFactory.getStopAgentFault(agent, runId).doTolerance();
			return null;
		}
	}

	private String buildBaseUrl(Agent agent) {
		return agent.getHostName() + ":" + agent.getPort();
	}

	public AgentBehaviorsBriefModel behaviorsBrief(Agent agent, UUID runId) {
		try {
			HttpResponse httpResponse = this.getHttpRequester().sendGet(
					buildBaseUrl(agent) + "/test/behaviorsBrief/"
							+ runId.toString(), null, null);
			if (HttpRequester.isInvalidResponse(httpResponse)) {
				return null;
			}
			logger.info(httpResponse.getContent());
			return (AgentBehaviorsBriefModel) MarshalHelper.unmarshal(
					AgentBehaviorsBriefModel.class, httpResponse.getContent());
		} catch (Exception e) {
			logger.error(ExceptionLog.getStackTrace(e) + " When behaviorsBrief");
			FaultTolerenceFactory.getBriefFaultTolerance(agent).doTolerance();
			return null;
		}
	}

	public AgentPageBriefModel pageBrief(Agent agent, UUID runId, int pageId) {
		try {
			HttpResponse httpResponse = this.getHttpRequester().sendGet(
					buildBaseUrl(agent) + "/test/pageBrief/" + runId.toString()
							+ "/" + pageId, null, null);
			if (HttpRequester.isInvalidResponse(httpResponse)) {
				return null;
			}

			return (AgentPageBriefModel) MarshalHelper.unmarshal(
					AgentPageBriefModel.class, httpResponse.getContent());
		} catch (Exception e) {
			logger.error(ExceptionLog.getStackTrace(e));
			FaultTolerenceFactory.getBriefFaultTolerance(agent).doTolerance();
			return null;
		}
	}

	public AgentPagesBriefModel pagesBrief(Agent agent, UUID runId) {
		try {
			HttpResponse httpResponse = this.getHttpRequester().sendGet(
					buildBaseUrl(agent) + "/test/pagesBrief/"
							+ runId.toString(), null, null);
			if (HttpRequester.isInvalidResponse(httpResponse)) {
				return null;
			}
			logger.info(httpResponse.getContent());
			return (AgentPagesBriefModel) MarshalHelper.unmarshal(
					AgentPagesBriefModel.class, httpResponse.getContent());
		} catch (Exception e) {
			logger.error(ExceptionLog.getStackTrace(e));
			FaultTolerenceFactory.getBriefFaultTolerance(agent).doTolerance();
			return null;
		}

	}

	public ServerStatusModel askLiving(String hostName, int port) {
		try {
			HttpResponse httpResponse = this.getHttpRequester().sendGet(
					hostName + ":" + port + "/", null, null);
			if (HttpRequester.isInvalidResponse(httpResponse)) {
				return null;
			}
			return (ServerStatusModel) MarshalHelper.unmarshal(
					ServerStatusModel.class, httpResponse.getContent());
		} catch (Exception e) {
			logger.error("agent :" + hostName + " break down!");
			return null;
		}
	}
}
