package org.bench4q.master.scriptrecord.httpcapture;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ConnectException;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.Socket;
import java.net.SocketException;
import java.net.URL;

import org.apache.log4j.Logger;
import org.bench4q.master.exception.ExceptionLog;

public class RequestHandler implements Runnable {
	private static final Logger log = Logger.getLogger(RequestHandler.class);
	private Config config = Config.getConfig();
	private InputStream clientIn;
	private InputStream serverIn;
	private OutputStream clientOut;
	private OutputStream serverOut;
	private HttpRequestHeader header;
	private ProxyServer proxyServer;
	private Socket clientSocket;
	private Socket serverSocket;
	private ByteArrayOutputStream buffer;
	private static Object mutex = new Object();

	public RequestHandler(ProxyServer proxyServer, Socket s) {
		assert (s != null);
		this.clientSocket = s;
		this.buffer = new ByteArrayOutputStream();
		this.proxyServer = proxyServer;
	}

	private void initClientServerConnections(Socket s) throws Throwable {
		this.clientIn = new BufferedInputStream(s.getInputStream());
		this.clientOut = s.getOutputStream();
		try {
			this.header = new HttpRequestHeader(this.clientIn);
		} catch (IOException e) {
			log.error("truncated request from browser: "
					+ ExceptionLog.getStackTrace(e).toString()
					+ this.header.url);
			System.out.println(this.header.url);
			throw new Utils.SilentException();
		}
		if (!this.header.url.startsWith("http"))
			throw new Utils.UserException(
					"Bench4Q only supports the HTTP protocol.");
		URL url = new URL(this.header.url);

		Config.ProxySettings proxy = this.config.getProxySettings();
		int port;
		String host;
		if (proxy != null) {
			host = proxy.host;
			port = proxy.port;
		} else {
			host = url.getHost();
			port = url.getPort();
		}

		if (port < 1)
			port = 80;
		try {
			this.serverSocket = new Socket(InetAddress.getByName(host), port);
		} catch (ConnectException e) {
			String msg = "Cannot connect to ";
			if (proxy != null)
				msg = msg + "proxy server ";
			msg = msg + host;
			if (port != 80)
				msg = msg + " on port " + Integer.toString(port);
			throw new Utils.UserException(msg + ".");
		}
		try {
			this.serverIn = this.serverSocket.getInputStream();
			this.serverOut = this.serverSocket.getOutputStream();
		} catch (Throwable e) {
			this.serverSocket.close();
			this.serverSocket = null;
			throw e;
		}
	}

	private String stripProxyInfoFromRequestHeader()
			throws MalformedURLException {
		String res = "";
		String origUrl = this.header.url;
		URL url = new URL(origUrl);
		this.header.url = url.getFile();
		res = this.header.toString();
		this.header.url = origUrl;
		return res;
	}

	public void run() {
		try {
			try {
				int rs = this.clientSocket.getReceiveBufferSize();
				int ss = this.clientSocket.getSendBufferSize();
				int BUF_SIZE = rs < ss ? ss : rs;

				byte[] buf = new byte[BUF_SIZE];

				initClientServerConnections(this.clientSocket);
				String headerStr;
				if (this.config.getProxySettings() == null)
					headerStr = stripProxyInfoFromRequestHeader();
				else
					headerStr = this.header.toString();
				log.trace("read request header");

				byte[] bytes = headerStr.getBytes();
				this.serverOut.write(bytes, 0, bytes.length);

				log.trace("wrote request header");
				byte[] requestBody;
				if (this.header.contentLength > 0) {
					this.buffer.reset();
					int len = 0;
					int num = 0;
					while (num < this.header.contentLength) {
						try {
							len = readRequestBody(buf);
						} catch (SocketException e) {
							log.info("truncated request from browser: "
									+ e.getMessage());
							throw new Utils.SilentException();
						}
						if (len == 0 || len == -1) {
							System.out.println("Read " + num
									+ "byte from requestBody");
							break;
						}
						log.trace("read " + Integer.toString(len) + " bytes");
						this.serverOut.write(buf, 0, len);
						this.buffer.write(buf, 0, len);
						log.trace("wrote " + Integer.toString(len) + " bytes");
						num += len;
					}
					requestBody = this.buffer.toByteArray();
					log.trace("transferred rest of request body");
				} else {
					requestBody = new byte[0];
					log.trace("no request body");
				}

				this.clientSocket.shutdownInput();

				this.serverSocket.shutdownOutput();

				synchronized (mutex) {
					this.proxyServer.processRequest(this.header, requestBody);
					log.trace("processed request");

					this.buffer.reset();
					int len;
					while ((len = this.serverIn.read(buf, 0, buf.length)) >= 0) {
						log.trace("read " + Integer.toString(len));
						try {
							this.clientOut.write(buf, 0, len);
						} catch (SocketException e) {
							log.info("browser stopped listening: "
									+ e.getMessage() + this.buffer.toString());
							throw new Utils.SilentException();
						}
						this.buffer.write(buf, 0, len);
					}
					log.trace("transferred response len:" + len);

					this.proxyServer.processResponse(this.header, requestBody,
							this.buffer.toByteArray());
					log.trace("processed response");
				}

			} catch (Exception e) {
				log.error(ExceptionLog.getStackTrace(e));
			} finally {
				if (this.serverSocket != null) {
					this.serverSocket.close();
					log.trace("closed server socket");
				}

				this.clientSocket.close();
				log.trace("closed client socket");
			}
			if (this.serverSocket != null) {
				this.serverSocket.close();
				log.trace("closed server socket");
			}

			this.clientSocket.close();
			log.trace("closed client socket");
		} catch (Throwable localThrowable) {
			log.error(ExceptionLog.getStackTrace(localThrowable));
		}
	}

	private int readRequestBody(byte[] buf) throws IOException {
		int alreadyReadLength = 0;
		int ch = -1;
		while (alreadyReadLength < this.header.contentLength) {
			if ((ch = this.clientIn.read()) != -1) {
				buf[alreadyReadLength] = (byte) ch;
			}
			alreadyReadLength++;
		}
		return alreadyReadLength;
	}
}
